/**
  ******************************************************************************
  * @file    mldr149secLoader.cpp
  * @author  Milandr Application Team
  * @version V1.0.0
  * @date    26/10/2020
  * @brief   MLDR149 security core S/N and gamma loading implementation
  ******************************************************************************
  * <br><br>
  *
  * THE PRESENT FIRMWARE WHICH IS FOR GUIDANCE ONLY AIMS AT PROVIDING CUSTOMERS
  * WITH CODING INFORMATION REGARDING THEIR PRODUCTS IN ORDER FOR THEM TO SAVE
  * TIME. AS A RESULT, MILANDR SHALL NOT BE HELD LIABLE FOR ANY
  * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
  * FROM THE CONTENT OF SUCH FIRMWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
  * CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
  *
  * <h2><center>&copy; COPYRIGHT 2020 Milandr</center></h2>
  */

#include <iostream>
#include <Windows.h>

#include "crc.h"
#include "rmd160.h"
#include "types.h"

typedef void* UART;

UART uartInit()
{
    HANDLE hSerial;
    LPCTSTR sPortName = L"\\\\.\\COM42";
    uint32_t err = 0;
    
    hSerial = CreateFile(sPortName, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL| FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH, 0);
    if (hSerial == INVALID_HANDLE_VALUE)
    {
        err = GetLastError();
        if (err == ERROR_FILE_NOT_FOUND)
        {
            std::cout << "serial port does not exist.\n";
        }
        std::cout << "some other error occurred.\n";
        return INVALID_HANDLE_VALUE;
    }
    
    DCB dcbSerialParams = { 0 };
    dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
    if (!GetCommState(hSerial, &dcbSerialParams))
    {
        std::cout << "getting state error\n";
        return INVALID_HANDLE_VALUE;
    }
    dcbSerialParams.BaudRate = /*512000;*//*CBR_128000*/CBR_9600/*CBR_115200*/;
    dcbSerialParams.ByteSize = 8;
    dcbSerialParams.StopBits = ONESTOPBIT;
    dcbSerialParams.Parity = NOPARITY;
    if (!SetCommState(hSerial, &dcbSerialParams))
    {
        std::cout << "error setting serial port state\n";
        return INVALID_HANDLE_VALUE;
    }
    
    return hSerial;
}

int uartRead(UART uart, uint8_t* data, uint32_t sz, uint32_t timeout5)
{
    DWORD actual = 0;
    uint32_t i = 0;

	printf("uartRead  %d: ", sz);
	while (0 != sz)
    {
        ReadFile(uart, &data[i], 1, &actual, NULL);
		for (uint32_t t = 0; t < actual; t++)
			printf("%02X ", data[i + t]);
		i += actual;
        sz -= actual;

        if (timeout5 == 0 && 0 != sz)
        {
            return 0xFF;
        }
        --timeout5;
        Sleep(5);
    }
	printf("\n");
	return 0;
}

bool uartWrite(UART uart, uint8_t* data, uint32_t sz)
{
    DWORD actual;
	printf("uartWrite %d: ", sz);

	actual = 0;
	DWORD s;
	uint32_t err = 0;
	for (uint32_t t = 0; t < sz; t++) {
		WriteFile(uart, &data[t], 1, &s, NULL);
		printf("%02X ", data[t]);
		actual += s;
		Sleep(1);
	}

	printf("\n");

    if (actual != sz)
    {
        return false;
    }
    return true;
}

void uartClose(UART uart)
{
    CloseHandle(uart);
}

void genSN(uint32_t* sn)
{
    sn[0] = 0xFEDCBA98;
    sn[1] = 0x76543210;
}

void genGamma(uint32_t master, uint32_t* sn, uint8_t* gamma, uint32_t sz)
{
    memset(gamma, 42, sz);
}

void main()
{
    uint32_t status = 0;
    UART     uart = NULL;
    uint32_t sn[2] = {0};
    uint32_t crc = 0;
    const uint32_t gammaDesiredSz = 16364;
    uint8_t  gamma[gammaDesiredSz] = { 0 };
    uint32_t master = 0x45454545;

	uint32_t OTP[4] = { 0x0000A55A, 0xFEDCBA98, 0x76543210, 0xd76cc5f2};
	crc = 0;
	crc = crc32(crc, (uint8_t*)OTP, 3*4);
	crc = 0;

    uart = uartInit();
    status = 0;
    for (;;)
    {
        crc = 0;
        status >>= 8;
        uartRead(uart, &((uint8_t*)&status)[3], 1, 0xffffffff); // get status
        switch (status)
        {
        case CM0_PREINIT_READY:
            printf("Preinit start.\n");
            genSN(sn);
            crc = crc32(crc, (uint8_t*)sn, 8);
            uartWrite(uart, (uint8_t*)sn, 8);
            uartWrite(uart, (uint8_t*)&crc, 4);
            uartRead(uart, (uint8_t*)&status, 4, 0xffffffff); // get status
            if (status == CM0_PREINIT_DONE)
            {
                printf("Preinit done.\n");
            }
            else
            {
                printf("Preinit fail.\n");
            }
            break;
        case CM0_INIT_READY:
            printf("Init start.\n");
            status = CM0_GAMMA_READY;
            uartWrite(uart, (uint8_t*)&status, 4);
            uartRead(uart, (uint8_t*)sn, 8, 0xffffffff);

            genGamma(master, sn, gamma, gammaDesiredSz); // generate
            uartWrite(uart, gamma, gammaDesiredSz);      // send to zmk
            crc = crc32(crc, gamma, gammaDesiredSz);     // calc next portion crc
            uartWrite(uart, (uint8_t*)&crc, 4);
            uartRead(uart, (uint8_t*)&status, 4, 0xffffffff);        // get status
            if (status == CM0_INIT_DONE)
            {
                printf("Init done.\n");
            }
            else
            {
                printf("Init fail.\n");
            }
            break;
        case 0x8FA55AF1:
            printf("Working.\n");
        }
    }
}


